
//////////////////////////////////////////////////////////////////////////////////
// This program is free software: you can redistribute it and/or modify         //
// it under the terms of the version 3 of the GNU General Public License        //
// as published by the Free Software Foundation.                                //
//                                                                              //
// This program is distributed in the hope that it will be useful, but          //
// WITHOUT ANY WARRANTY; without even the implied warranty of                   //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU             //
// General Public License for more details.                                     //
//                                                                              //
// You should have received a copy of the GNU General Public License            //
// along with this program. If not, see <http://www.gnu.org/licenses/>.         //
//                                                                              //
// Written by Francois Fleuret                                                  //
// (C) Ecole Polytechnique Federale de Lausanne                                 //
// Contact <francois.fleuret@epfl.ch> for comments & bug reports                //
//////////////////////////////////////////////////////////////////////////////////

// $Id: fastentropy.cc,v 1.4 2007-08-23 08:36:50 fleuret Exp $

using namespace std;

#include <iostream>
#include <cmath>
#include <stdio.h>
#include <stdlib.h>

#include "misc.h"
#include "fastentropy.h"

int fe_nb_bits[65536];
double fe_logn[65536], fe_nlogn[65536];

#ifdef DEBUG
int fe_was_initialized = 0;
#endif

inline void fe_and(int n, uint32_t *a, uint32_t *b, uint32_t *ab) {
  for(int k = 0; k < (n+31)/32; k++) *ab++ = *a++ & *b++;
}

inline void fe_and_not(int n, uint32_t *a, uint32_t *b, uint32_t *ab) {
  for(int k = 0; k < (n+31)/32; k++) *ab++ = *a++ & ~*b++;
}

inline void fe_and_not_not(int n, uint32_t *a, uint32_t *b, uint32_t *ab) {
  for(int k = 0; k < (n+31)/32; k++) *ab++ = ~*a++ & ~*b++;
}

inline int fe_count(int n, uint32_t *a) {
  uint16_t *aa = (uint16_t *) a;
  int t = 0;
  for(int k = 0; k < n/16; k++) t += fe_nb_bits[*aa++];
  if(n%16 > 0) t += fe_nb_bits[(65535 >> (16-(n%16))) & *aa];
  return t;
}

inline double fe_entropy(int n, uint32_t *a) {
  int n1 = fe_count(n, a);
  return fe_logn[n] - (fe_nlogn[n1] + fe_nlogn[n-n1])/double(n);
}

inline double fe_entropy_couple(int n, uint32_t *a, uint32_t *b) {
  int n11 = fe_count_and(n, a, b);
  int n10 = fe_count_and_not(n, a, b);
  int n01 = fe_count_and_not(n, b, a);
  int n00 = fe_count_and_not_not(n, a, b);

  return fe_logn[n] - (fe_nlogn[n00] + fe_nlogn[n01] + fe_nlogn[n10] + fe_nlogn[n11])/double(n);
}

void fe_init_tables() {
#ifdef DEBUG
  fe_was_initialized = 1;
#endif
  for(int i = 0; i < 65536; i++) {
    if(i == 0) { fe_logn[i] = 0.0; fe_nlogn[i] = 0.0; }
    else { fe_logn[i] = log(double(i)); fe_nlogn[i] = double(i) * log(double(i)); }
    int n = 0;
    for(int j = 0; j < 16; j++) if(i & (1 << j)) n++;
    fe_nb_bits[i] = n;
  }
}

void fe_selection_cmim(int nb_samples,
                       int nb_tests, uint32_t **x, uint32_t *y,
                       int nb_selected, int *selected) {

#ifdef DEBUG
  if(!fe_was_initialized) {
    cerr << "fe_init_tables() was not called!\n";
    abort();
  }
#endif

  if(nb_samples > 65535) {
    cerr << "Too many pictures, the nlogn table is too small.\n";
    exit(1);
  }

  double s[nb_tests];
  double ch[nb_tests];
  int m[nb_tests];

  double h = fe_entropy(nb_samples, y);

  for(int i = 0; i < nb_tests; i++) {
    ch[i] = fe_entropy_couple(nb_samples, y, x[i]) - fe_entropy(nb_samples, x[i]);
    s[i] = h - ch[i];
    m[i] = 0;
  }

  for(int n = 0; n < nb_selected; n++) {
    double best_s = 0;
    for(int i = 0; i < nb_tests; i++) {
      if(s[i] > best_s) {
        int nb_uint32 = (nb_samples+31)/32;
        if(m[i] < n) {
          uint32_t z00[nb_uint32], z01[nb_uint32], z10[nb_uint32], z11[nb_uint32];
          fe_and(nb_samples, x[i], y, z11);
          fe_and_not(nb_samples, x[i], y, z10);
          fe_and_not(nb_samples, y, x[i], z01);
          fe_and_not_not(nb_samples, x[i], y, z00);
          while(s[i] > best_s && m[i] < n) {
            double h_y_xi_xmi = fe_logn[nb_samples] -
              (  fe_nlogn[fe_count_and    (nb_samples, z11, x[selected[m[i]]])]
               + fe_nlogn[fe_count_and    (nb_samples, z10, x[selected[m[i]]])]
               + fe_nlogn[fe_count_and    (nb_samples, z01, x[selected[m[i]]])]
               + fe_nlogn[fe_count_and    (nb_samples, z00, x[selected[m[i]]])]
               + fe_nlogn[fe_count_and_not(nb_samples, z11, x[selected[m[i]]])]
               + fe_nlogn[fe_count_and_not(nb_samples, z10, x[selected[m[i]]])]
               + fe_nlogn[fe_count_and_not(nb_samples, z01, x[selected[m[i]]])]
               + fe_nlogn[fe_count_and_not(nb_samples, z00, x[selected[m[i]]])])/double(nb_samples);
            double h_xi_xmi = fe_entropy_couple(nb_samples, x[i], x[selected[m[i]]]);
            double ss = ch[selected[m[i]]] - (h_y_xi_xmi - h_xi_xmi);
            if(ss < s[i]) s[i] = ss;
            m[i]++;
          }
        }
        if(s[i] > best_s) {
          best_s = s[i];
          selected[n] = i;
        }
      }
    }
  }
}

void fe_selection_mim(int nb_samples,
                      int nb_tests, uint32_t **x, uint32_t *y,
                      int nb_selected, int *selected) {

#ifdef DEBUG
  if(!fe_was_initialized) {
    cerr << "fe_init_tables() was not called!\n";
    abort();
  }
#endif

  if(nb_samples > 65535) {
    cerr << "Too many pictures, the nlogn table is too small.\n";
    exit(1);
  }

  Couple tmp[nb_tests];
  double h = fe_entropy(nb_samples, y);

  for(int i = 0; i < nb_tests; i++) {
    tmp[i].index = i;
    tmp[i].value = fe_entropy_couple(nb_samples, y, x[i]) - h - fe_entropy(nb_samples, x[i]);
  }

  qsort(tmp, nb_tests, sizeof(Couple), compare_couple);

  // Here we have the features sorted according to their mutual
  // information with the class to predict

  for(int n = 0; n < nb_selected; n++) selected[n] = tmp[n].index;
}
